@@ -1,9 +1,11 @@
 use serde_json::Value;
 use std::usize;
-use swc_common::{Spanned, DUMMY_SP};
+use swc_common::{util::take::Take, Spanned, DUMMY_SP};
 use swc_ecma_ast::*;
+use swc_ecma_transforms_base::perf::Parallel;
+use swc_ecma_transforms_macros::parallel;
 use swc_ecma_utils::{calc_literal_cost, member_expr, ExprFactory};
-use swc_ecma_visit::{noop_fold_type, Fold, FoldWith};
+use swc_ecma_visit::{as_folder, noop_visit_mut_type, Fold, VisitMut, VisitMutWith};
 
 /// Transform to optimize performance of literals.
 ///
@@ -28,39 +30,49 @@ use swc_ecma_visit::{noop_fold_type, Fold, FoldWith};
 ///
 /// See https://github.com/swc-project/swc/issues/409
 pub fn json_parse(min_cost: usize) -> impl Fold {
-    JsonParse { min_cost }
+    as_folder(JsonParse { min_cost })
 }
 
 struct JsonParse {
     pub min_cost: usize,
 }
 
+impl Parallel for JsonParse {
+    fn create(&self) -> Self {
+        JsonParse {
+            min_cost: self.min_cost,
+        }
+    }
+
+    fn merge(&mut self, _: Self) {}
+}
+
 impl Default for JsonParse {
     fn default() -> Self {
         JsonParse { min_cost: 1024 }
     }
 }
 
-/// TODO: VisitMut
-impl Fold for JsonParse {
-    noop_fold_type!();
+#[parallel]
+impl VisitMut for JsonParse {
+    noop_visit_mut_type!();
 
     /// Handles parent expressions before child expressions.
-    fn fold_expr(&mut self, e: Expr) -> Expr {
+    fn visit_mut_expr(&mut self, expr: &mut Expr) {
         if self.min_cost == usize::MAX {
-            return e;
+            return;
         }
 
-        let e = match e {
+        let e = match expr {
             Expr::Array(..) | Expr::Object(..) => {
-                let (is_lit, cost) = calc_literal_cost(&e, false);
+                let (is_lit, cost) = calc_literal_cost(&*expr, false);
                 if is_lit && cost >= self.min_cost {
-                    return Expr::Call(CallExpr {
-                        span: e.span(),
+                    *expr = Expr::Call(CallExpr {
+                        span: expr.span(),
                         callee: member_expr!(DUMMY_SP, JSON.parse).as_callee(),
                         args: vec![Lit::Str(Str {
                             span: DUMMY_SP,
-                            value: serde_json::to_string(&jsonify(e))
+                            value: serde_json::to_string(&jsonify(expr.take()))
                                 .unwrap_or_else(|err| {
                                     unreachable!(
                                         "failed to serialize serde_json::Value as json: {}",
@@ -74,14 +86,15 @@ impl Fold for JsonParse {
                         .as_arg()],
                         type_args: Default::default(),
                     });
+                    return;
                 }
 
-                e
+                expr
             }
-            _ => e,
+            _ => expr,
         };
 
-        e.fold_children_with(self)
+        e.visit_mut_children_with(self)
     }
 }
 
@@ -135,7 +148,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         simple_object,
         "let a = {b: 'foo'}",
         r#"let a = JSON.parse('{"b":"foo"}')"#
@@ -143,7 +156,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         simple_arr,
         "let a = ['foo']",
         r#"let a = JSON.parse('["foo"]')"#
@@ -151,7 +164,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         empty_object,
         "const a = {};",
         r#"const a = JSON.parse('{}');"#
@@ -159,7 +172,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 15 },
+        |_| json_parse(15),
         min_cost_15,
         "const a = { b: 1, c: 2 };",
         "const a = { b: 1, c: 2 };"
@@ -167,7 +180,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         min_cost_0,
         "const a = { b: 1, c: 2 };",
         r#"const a = JSON.parse('{"b":1,"c":2}');"#
@@ -175,7 +188,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         spread,
         "const a = { ...a, b: 1 };",
         "const a = { ...a, b: 1 };"
@@ -183,7 +196,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         object_method,
         "const a = {
         method(arg) {
@@ -201,7 +214,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         computed_property,
         r#"const a = { b : "b_val", ["c"]: "c_val" };"#,
         r#"const a = { b : "b_val", ["c"]: "c_val" };"#
@@ -209,7 +222,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         invalid_numeric_key,
         r#"const a ={ 77777777777777777.1: "foo" };"#,
         r#"const a = JSON.parse('{"77777777777777780":"foo"}');"#
@@ -217,7 +230,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         string,
         r#"const a = { b: "b_val" };"#,
         r#"const a = JSON.parse('{"b":"b_val"}');"#
@@ -225,7 +238,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         string_single_quote_1,
         r#"const a = { b: "'abc'" };"#,
         r#"const a = JSON.parse('{"b":"\'abc\'"}');"#,
@@ -234,7 +247,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         string_single_quote_2,
         r#"const a = { b: "ab\'c" };"#,
         r#"const a = JSON.parse('{"b":"ab\'c"}');"#,
@@ -243,7 +256,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         number,
         "const a = { b: 1 };",
         r#"const a = JSON.parse('{"b":1}');"#
@@ -251,7 +264,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         null,
         "const a = { b: null };",
         r#"const a = JSON.parse('{"b":null}');"#
@@ -259,7 +272,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         boolean,
         "const a = { b: false };",
         r#"const a = JSON.parse('{"b":false}');"#
@@ -267,7 +280,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         array,
         "const a = { b: [1, 'b_val', null] };",
         r#"const a = JSON.parse('{"b":[1,"b_val",null]}');"#
@@ -275,7 +288,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         nested_array,
         "const a = { b: [1, ['b_val', { a: 1 }], null] };",
         r#"const a = JSON.parse('{"b":[1,["b_val",{"a":1}],null]}');"#
@@ -283,7 +296,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         object,
         "const a = { b: { c: 1 } };",
         r#"const a = JSON.parse('{"b":{"c":1}}');"#
@@ -291,7 +304,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |_| JsonParse { min_cost: 0 },
+        |_| json_parse(0),
         object_numeric_keys,
         r#"const a = { 1: "123", 23: 45, b: "b_val" };"#,
         r#"const a = JSON.parse('{"1":"123","23":45,"b":"b_val"}');"#
